using System;
using System.IO;
using System.Linq;
using System.Text;
using gaemstone.Client.Utility;
using gaemstone.ECS;
using gaemstone.ECS.Utility;
using gaemstone.Flecs;
using ImGuiNET;
using Silk.NET.Input;
using Silk.NET.OpenGL.Extensions.ImGui;
using static gaemstone.Client.Components.InputComponents;
using static gaemstone.Client.Systems.InputManager;
using static gaemstone.Client.Systems.Windowing;

namespace gaemstone.Client.Systems;

[Module]
[DependsOn<gaemstone.Client.Components.InputComponents>]
[DependsOn<gaemstone.Client.Systems.InputManager>]
[DependsOn<gaemstone.Client.Systems.Windowing>]
public partial class ImGuiManager
{
	[Entity, Add<Pipeline.Phase>]
	[DependsOn<Pipeline.OnLoad>]
	public struct ImGuiUpdatePhase {  }

	[Entity, Add<Pipeline.Phase>]
	[DependsOn<Pipeline.OnStore>]
	public struct ImGuiRenderPhase {  }

	[Singleton]
	public class ImGuiData
	{
		public ImGuiController Controller { get; }
		internal ImGuiData(ImGuiController controller) => Controller = controller;
	}

	private unsafe static ImFontPtr AddFontFromResources(
		ImGuiIOPtr io, string name, int size, Action<ImFontConfigPtr>? cfgAction = null,
		float? offset = null, float? advance = null, bool merge = false,
		(int Min, int Max)[]? ranges = null)
	{
		// TODO: FontConfig can be freed at the end of this method.
		//       Unfortunately, data has to stick around until font atlas is built.
		var cfg = new ImFontConfigPtr(ImGuiNative.ImFontConfig_ImFontConfig());

		if (offset is float o) cfg.GlyphOffset = new(0, o);
		if (advance is float a) cfg.GlyphMinAdvanceX = cfg.GlyphMaxAdvanceX = a;
		if (merge) cfg.MergeMode = merge;

		// Set cfg.Name so the font has a nice display name in ImGui.
		var fullName = $"{name.Replace('.', ' ')} {size}px";
		Encoding.UTF8.GetBytes(fullName, new Span<byte>(cfg.Name.Data, cfg.Name.Count));

		// If glyph ranges are specified, allocate unmanaged heap memory for them.
		if (ranges != null) {
			var rangesSpan = GlobalHeapAllocator.Instance.Allocate<char>(ranges.Length * 2 + 1);
			for (var i = 0; i < ranges.Length; i++) {
				rangesSpan[i * 2    ] = (char)ranges[i].Min;
				rangesSpan[i * 2 + 1] = (char)ranges[i].Max;
			}
			rangesSpan[^1] = default;
			fixed (void* rangesPtr = rangesSpan)
				cfg.GlyphRanges = (IntPtr)rangesPtr;
		}

		// Use cfgAction to allow changing other font configs.
		cfgAction?.Invoke(cfg);

		// Grab the stream for this font from the assembly's resources.
		var file = $"gaemstone.Client.Resources.{name}.ttf";
		using var stream = typeof(Resources).Assembly.GetManifestResourceStream(file)
			?? throw new InvalidOperationException($"Resource '{file}' was not found");

		using var memoryStream = new MemoryStream();
		stream.CopyTo(memoryStream); // Write font file from resources to memory stream.
		memoryStream.WriteByte(0);   // Add a NUL termination character.

		// Copy the data into unmanaged memory and pass it to ImGui.
		var fontData = memoryStream.ToArray();
		var fontDataSpan = GlobalHeapAllocator.Instance.AllocateCopy<byte>(fontData);
		fixed (byte* dataPtr = fontDataSpan)
			return io.Fonts.AddFontFromMemoryTTF((IntPtr)dataPtr, size, size, cfg);
	}

	[System]
	[DependsOn<Pipeline.OnLoad>]
	public static unsafe void Initialize<T>(World<T> world, GameWindow window, Canvas canvas,
			[Source<Input>] InputContext inputContext, Not<ImGuiData> _)
		=> world.Entity<ImGuiData>().Set(new ImGuiData(
			new(canvas.GL, window.Handle, inputContext.Value, () => {
				var io    = ImGui.GetIO();
				var style = ImGui.GetStyle();

				var fontSize = 16;
				void MergeIcons() => AddFontFromResources(io,
					"ForkAwesome", fontSize, advance: fontSize, merge: true,
					ranges: new[]{ (ForkAwesome.Min, ForkAwesome.Max) });
				AddFontFromResources(io, "OpenSans"       , fontSize, offset: -1); MergeIcons();
				AddFontFromResources(io, "OpenSans.Bold"  , fontSize, offset: -1); MergeIcons();
				AddFontFromResources(io, "OpenSans.Italic", fontSize, offset: -1); MergeIcons();

				// Do not save a settings or log file.
				io.NativePtr->IniFilename = null;
				io.NativePtr->LogFilename = null;

				io.BackendFlags |= ImGuiBackendFlags.HasMouseCursors
				                |  ImGuiBackendFlags.HasSetMousePos;
				io.ConfigFlags  |= ImGuiConfigFlags.NavEnableKeyboard
				                |  ImGuiConfigFlags.NavEnableSetMousePos;
				                // |  ImGuiConfigFlags.DockingEnable;
				io.ConfigWindowsResizeFromEdges = true;
				io.ConfigWindowsMoveFromTitleBarOnly = true;

				style.WindowRounding = style.ChildRounding
					= style.PopupRounding = style.TabRounding
					= style.FrameRounding = style.ScrollbarRounding = 6;

				style.WindowTitleAlign = new(0.5f, 0.5f);
				style.ColorButtonPosition = ImGuiDir.Left;
				style.IndentSpacing = 8;

				// Set up key mappings.
				foreach (var imguiKey in Enum.GetValues<ImGuiKey>()) {
					var name = imguiKey.ToString();

					// Adjust ImGuiKey enum names to match Silk.NET.Input.Key enum.
					if (name.StartsWith("_")) name = name.Replace("_", "Number");
					if (name.EndsWith("Arrow")) name = name[..^"Arrow".Length];
					if (name.EndsWith("Ctrl" )) name = name.Replace("Ctrl", "Control");

					if (!name.EndsWith("Bracket")) { // Leave "LeftBracket" and "RightBracket" alone.
						if (name.StartsWith("Left" )) name = name["Left" .Length..] + "Left";
						if (name.StartsWith("Right")) name = name["Right".Length..] + "Right";
					}

					if (Enum.TryParse<Key>(name, true, out var silkKey))
						io.KeyMap[(int)imguiKey] = (int)silkKey;
				}
			})));

	[System]
	[DependsOn<Pipeline.OnLoad>]
	public static void UpdateMouse<T>(World<T> world,
		[Source<Mouse>] MouseImpl impl, ImGuiData _)
	{
		var mouse  = impl.Value;
		var input  = world.Entity<Input>();
		var module = world.Entity<ImGuiManager>();
		var capturedBy = input.GetTargets<MouseInputCapturedBy>().FirstOrDefault();;
		var isCaptured = capturedBy.IsSome;
		// If another system has the mouse captured, don't do anything here.
		if (isCaptured && (capturedBy != module)) return;

		var IO = ImGui.GetIO();

		// Set the mouse position if ImGui wants to move it.
		if (IO.WantSetMousePos) mouse.Position = IO.MousePos;

		// Capture the mouse input it is above GUI elements.
		if (IO.WantCaptureMouse != isCaptured) {
			if (IO.WantCaptureMouse) input.Add   <MouseInputCapturedBy>(module);
			else                     input.Remove<MouseInputCapturedBy>(module);
		}

		var cursor = ImGui.GetMouseCursor();
		mouse.Cursor.CursorMode = (cursor == ImGuiMouseCursor.None)
			? CursorMode.Hidden : CursorMode.Normal;
		// TODO: Use additional cursors once Silk.NET supports GLFW 3.4.
		mouse.Cursor.StandardCursor = cursor switch {
			ImGuiMouseCursor.Arrow      => StandardCursor.Arrow,
			ImGuiMouseCursor.TextInput  => StandardCursor.IBeam,
			// ImGuiMouseCursor.ResizeAll  => StandardCursor.,
			ImGuiMouseCursor.ResizeNS   => StandardCursor.VResize,
			ImGuiMouseCursor.ResizeEW   => StandardCursor.HResize,
			// ImGuiMouseCursor.ResizeNESW => StandardCursor.,
			// ImGuiMouseCursor.ResizeNWSE => StandardCursor.,
			ImGuiMouseCursor.Hand       => StandardCursor.Hand,
			// ImGuiMouseCursor.NotAllowed => StandardCursor.,
			_ => StandardCursor.Default,
		};
	}

	[System]
	[DependsOn<ImGuiUpdatePhase>]
	public static void Update(TimeSpan delta, ImGuiData imgui)
		=> imgui.Controller.Update((float)delta.TotalSeconds);

	[System]
	[DependsOn<ImGuiRenderPhase>]
	public static void Render(ImGuiData imgui)
		=> imgui.Controller.Render();
}
